namespace PhysicsEngine 
{
    using System;
    using System.Data;
    using System.Collections;
    using Microsoft.DirectX;
    using Microsoft.DirectX.Direct3D;

    public interface RigidBodyList
    {
        int GetNumBodies();
        RigidBody GetBody(int i);
    }

    
    public class RigidBodyType
    {
        private string name;
        public string Name { get { return name; } }
        
        private string meshFileName;

        private Mesh mesh;
        private Material[] materials;
        private Texture[] textures;

        public Polyhedron polyhedron;

        public SDM sdm;

        public float volume;
        public Matrix3 inertiaTensor;
        public float radius;

        public RigidBodyType(string name, string meshFileName, Device device)
        {
            // TODO make this more general, so it doesn't rely on polyhedron, doesn't do the mesh loading, etc.
            
            this.name = name;
            this.meshFileName = meshFileName;

            /* load mesh from file, compute center of mass, moment
                 * of intertia, transform mesh so center of mass is 0
                 * and moments are on primary axis */
            ExtendedMaterial[] exMaterials;
            mesh = Mesh.FromFile(meshFileName, MeshFlags.SystemMemory, device, out exMaterials);
            textures = new Texture[exMaterials.Length]; 
            materials = new Material[exMaterials.Length]; 
            for ( int i = 0; i < exMaterials.Length; ++i) 
            { 
                if (exMaterials[i].TextureFilename != null ) 
                { 
                    string texturePath = 
                        System.IO.Path.Combine(System.IO.Path.GetDirectoryName(meshFileName), 
                        exMaterials[i].TextureFilename); 
                    textures[i] = TextureLoader.FromFile(device, texturePath); 
                } 
                materials[i] = exMaterials[i].Material3D; 
                materials[i].Ambient = materials[i].Diffuse; 
            } 


            polyhedron = new Polyhedron(mesh);
  
            Vector3 centerOfMass;
            polyhedron.ComputeMassProperties(1.0f, out volume, out centerOfMass, out inertiaTensor);

            // transform vertices in original mesh to make the center of mass (0,0,0)
            CustomVertex.PositionNormalTextured[] vertList = (CustomVertex.PositionNormalTextured[])mesh.LockVertexBuffer(typeof(CustomVertex.PositionNormalTextured), LockFlags.None, mesh.NumberVertices);
            for (int i = 0; i < vertList.Length; i++)
            {
                vertList[i].SetPosition(vertList[i].GetPosition() - centerOfMass);
            }
            mesh.UnlockVertexBuffer();
           
            // transform the polyhedron as well
            polyhedron.Transform(Matrix.Translation(-1 * centerOfMass));

            //sdm = new BruteForceSDM(polyhedron);
            sdm = new GridSDM(polyhedron, 0.1f);

            radius = 0;
            foreach (Polyhedron.Vertex v in polyhedron.verts)
                radius = Math.Max(radius, v.pos.Length());
        }

        public void Render(Device device)
        {
            for ( int i = 0; i < materials.Length; ++i) 
            { 
                if (textures[i] != null ) 
                { 
                    device.SetTexture(0, textures[i]); 
                } 
                device.Material = materials[i]; 
                mesh.DrawSubset(i); 
            } 
        }
    }

    public class RigidBody
    {
        public class State
        {
            public Vector3 pos;
            public Vector3 vel;
            public Quaternion orientation;
            public Vector3 rotation;
            public State(RigidBody b)
            {
                pos = b.pos;
                vel = b.vel;
                orientation = b.orientation;
                rotation = b.rotation;
            }
        }


        private string name;
        private RigidBodyType type;
        private float bounciness;
        private float friction;
        private Vector3 pos;
        private Vector3 vel;
        private Quaternion orientation;
        private Vector3 rotation;
        private bool moved;
        private Matrix3 orientationMatrix;
        private Matrix3 inverseOrientationMatrix;
        private Matrix3 inverseWorldInertiaMatrix;
        private float mass;
        private Matrix3 inverseInertiaTensor;
        private Matrix3 inverseMassMatrix;
        
        public int level;
        public bool levelSet;
        public ArrayList collisions;

        // used in shock propagation
        public bool frozen;
    
        public string Name
        {
            get { return name; }
        }

        public Vector3 Pos 
        {
            get { return pos; }
            set { pos = value; moved = true; }
        }

        public Vector3 Vel 
        {
            get { return vel; }
            set { vel = value; }
        }

        public Quaternion Orientation 
        {
            get { return orientation; }
            set 
            { 
                // XXX inefficient (but safer) to renormalize every time
                if (value.LengthSq() == 0)
                    orientation = Quaternion.Identity;
                 else
                    orientation = Quaternion.Normalize(value);
                

                //sanity check
                if (float.IsNaN(orientation.LengthSq()))
                {
                    Console.Out.WriteLine("sanity check!");
                    while (true) {}
                }
                moved = true;
            }
        }

        public Vector3 Rotation 
        {
            get { return rotation; }
            set 
            { 
                // sanity check
                rotation = value;
                if (float.IsNaN(rotation.LengthSq()))
                {
                    Console.Out.WriteLine("sanity check!");
                    while (true) {}
                }
            }
        }

        public Matrix3 OrientationMatrix
        {
            get
            {
                if (moved)
                    RebuildMatrices();
                return orientationMatrix;
            }
        }

        public Matrix3 InverseOrientationMatrix
        {
            get
            {
                if (moved)
                    RebuildMatrices();
                return inverseOrientationMatrix;
            }
        }

        public Matrix3 InverseWorldInertiaMatrix
        {
            get
            {
                if (moved)
                    RebuildMatrices();
                return inverseWorldInertiaMatrix;
            }
        }

        private void RebuildMatrices()
        {
            moved = false;
            orientationMatrix = Matrix3.FromMatrix4(Matrix.RotationQuaternion(orientation));
            inverseOrientationMatrix = orientationMatrix.Transpose();
            inverseWorldInertiaMatrix = orientationMatrix * inverseInertiaTensor * inverseOrientationMatrix;
        }

        /* computes the velocity in world space of the
                 * point x in object space */
        public Vector3 WorldVelocity(Vector3 x)
        {
            // v_world = v + cross(omega, R*x_body)
            return Vel + Vector3.Cross(Rotation, x - Pos);
        }

        public Vector3 PointWorldToBody(Vector3 p)
        {
            return InverseOrientationMatrix * (p - Pos);
        }

        public Vector3 PointBodyToWorld(Vector3 p)
        {
            return OrientationMatrix * p + Pos;
        }

        public Vector3 VectorWorldToBody(Vector3 v)
        {
            return InverseOrientationMatrix * v;
        }

        public Vector3 VectorBodyToWorld(Vector3 v)
        {
            return OrientationMatrix * v;
        }



        public RigidBody(string name,
            RigidBodyType type,
            Vector3 pos,
            Vector3 vel,
            Quaternion orientation,
            Vector3 rotation,
            float bounciness,
            float friction,
            float density)
        {
            this.name = name;
            this.type = type;
            Pos = pos;
            Vel = vel;
            Orientation = orientation;
            Rotation = rotation;
            this.bounciness = bounciness;
            this.friction = friction;
            this.collisions = new ArrayList();
            this.mass = density * type.volume;
            inverseInertiaTensor = (density * type.inertiaTensor).Inverse();
            inverseMassMatrix = Matrix3.Identity * (1/mass);
        }

        public void SetState(State state)
        {
            Pos = state.pos;
            Orientation = state.orientation;
            Vel = state.vel;
            Rotation = state.rotation;
        }
        
        public RigidBodyType Type
        {
            get { return type; }
        }

        public float Bounciness
        {
            get { return bounciness; }
        }

        public float Friction
        {
            get { return friction; }
        }

        public float Mass
        {
            get { return mass; }
        }

        // evaluates the SDM at point p in world space
        public SDM.Value EvaluateSDM(Vector3 p, bool compute_normal)
        {
            return type.sdm.Evaluate(p, compute_normal);
        }

        /* computes the K matrix given a point x in world space
             * (used in collision response) */
        public Matrix3 ComputeK(Vector3 x)
        {
            if (frozen || mass == float.PositiveInfinity)
                return Matrix3.Zero;

            Matrix3 Rx = Matrix3.CrossProductMatrix(x - Pos);
            return inverseMassMatrix - Rx * InverseWorldInertiaMatrix * Rx;
        }

        // applies an impulse J at point x (both in world coords) to the body
        public void ApplyImpulse(Vector3 x, Vector3 J)
        {
            if (frozen || mass == float.PositiveInfinity) return;
            
            if (!(-1e8 < J.X && J.X < 1e8 && -1e8 < J.Y && J.Y < 1e8 && -1e8 < J.Z && J.Z < 1e8))
            {
                Console.Out.WriteLine("sanity check!");
                while (true) {}
            }

            Vector3 Rx = x - Pos;
            Vector3 vOld = Vel;
            Vector3 vNew = vOld +  J * (1 / mass);
            Vector3 wOld = Rotation;
            Vector3 wNew = wOld + InverseWorldInertiaMatrix * Vector3.Cross(Rx, J);
            
            Vel = vNew;
            Rotation = wNew;
        }

        public void Render(Device device)
        {
            device.Transform.World = Matrix3.ToMatrix4(OrientationMatrix) * Matrix.Translation(Pos);
            type.Render(device);
        }

        public void UpdatePos(float dt)
        {
            /* beware of NaNs and infinity */
            Pos += Vel * dt;
            float rotSpeed = Rotation.Length();
            if (rotSpeed != 0)
                Orientation *= Quaternion.RotationAxis(Rotation * (1/rotSpeed), rotSpeed * dt);
        }

        public void UpdateVel(float dt)
        {
            // integrate forces over dt
            
            // gravity (for non-infinite objects
            if (mass != float.PositiveInfinity)
                Vel += dt * new Vector3(0, -9.8f, 0);
        }
    }
}