namespace PhysicsEngine 
{
    using System;
    using System.Collections;
    using Microsoft.DirectX;
    using Microsoft.DirectX.Direct3D;

    public class WorldState
    {
        private RigidBodyList bodies;
        private int num_collision_steps;
        private int num_contact_steps;
        private bool shock_propagation;
        private bool edge_intersection;
        private bool collisions_before_velocity_update;

        public RigidBodyList BodyList 
        {
            get { return bodies; }
            set { bodies = value; }
        }

        public int NumCollisionSteps
        {
            get { return num_collision_steps; }
            set { num_collision_steps = value; }
        }

        public int NumContactSteps
        {
            get { return num_contact_steps; }
            set { num_contact_steps = value; }
        }

        public bool ShockPropagation
        {
            get { return shock_propagation; }
            set { shock_propagation = value; }
        }

        public bool EdgeIntersection
        {
            get { return edge_intersection; }
            set { edge_intersection = value; }
        }

        public bool CollisionsBeforeVelocityUpdate
        {
            get { return collisions_before_velocity_update; }
            set { collisions_before_velocity_update = value; }
        }

        public WorldState(RigidBodyList bodies, int num_collision_steps, int num_contact_steps, bool shock_propagation, bool edge_intersection, bool collisions_before_velocity_update)
        {
            this.bodies = bodies;
            this.num_collision_steps = num_collision_steps;
            this.num_contact_steps = num_contact_steps;
            this.shock_propagation = shock_propagation;
            this.edge_intersection = edge_intersection;
            this.collisions_before_velocity_update = collisions_before_velocity_update;
        }

        public void Render(Device device)
        {
            if (bodies == null) return;
            for (int i = 0; i < bodies.GetNumBodies(); i++)
            {
                bodies.GetBody(i).Render(device);
            }
        }

        private CollisionPair CheckCollision(RigidBody[] b)
        {
            // if both bodies have infinite mass ignore collision
            if (b[0].Mass == float.PositiveInfinity && b[1].Mass == float.PositiveInfinity)
                return null;
            
            // TODO: better macro collision detection than bounding spheres

            // check bounding spheres
            if ((b[0].Pos - b[1].Pos).LengthSq() > (b[0].Type.radius + b[1].Type.radius) * (b[0].Type.radius + b[1].Type.radius))
                return null;


            ArrayList[] cp = new ArrayList[2];
            cp[0] = new ArrayList();
            cp[1] = new ArrayList();
            
            for (int i = 0; i < 2; i++)
            {
                foreach (Polyhedron.Vertex v in b[i].Type.polyhedron.verts)
                {
                    Vector3 p = b[1-i].PointWorldToBody(b[i].PointBodyToWorld(v.pos));
                    if (b[1-i].EvaluateSDM(p, false).d <= 0)
                        cp[i].Add(v.pos);
                }

                if (edge_intersection)
                {
                    // find the deepest point of penetration along each edge
                    float resolution = 0.66f;
                    foreach (Polyhedron.Edge e in b[i].Type.polyhedron.edges)
                    {
                        Vector3 edge_vector = b[1-i].VectorWorldToBody(b[i].VectorBodyToWorld(e.v[1].pos - e.v[0].pos));
                        float edge_length = edge_vector.Length();
                        Vector3 inc = edge_vector * (resolution / edge_length);
                        Vector3 p = b[1-i].PointWorldToBody(b[i].PointBodyToWorld(e.v[0].pos));
                        int n = (int)(edge_length / resolution);
                        //Console.Out.WriteLine("n = {0}", n);    

                        SDM.Value bestCP = new SDM.Value(new Vector3(), new Vector3(), float.PositiveInfinity);
                        for (; n > 0; n--)
                        {
                            p += inc;
                            SDM.Value curCP = b[1-i].EvaluateSDM(p, false);
                            if (curCP.d < bestCP.d && curCP.d <= 0)
                                bestCP = curCP;
                        }
                        if (bestCP.d != float.PositiveInfinity)
                            cp[i].Add(b[i].PointWorldToBody(b[1-i].PointBodyToWorld(bestCP.p)));
                    }
                }
            }

            if (cp[0].Count > 0 || cp[1].Count > 0)
                return new CollisionPair(b, cp);
            else
                return null;
        }

        private void ProcessAllCollisions(float dt)
        {
            ArrayList collisionPairs = new ArrayList();
    
            // collision processing
            RigidBody.State[] oldState = new RigidBody.State[bodies.GetNumBodies()];
            for (int step = 0; step < num_collision_steps; step++)
            {
                // temporarily update vel then pos for each body
                for (int i = 0; i < bodies.GetNumBodies(); i++)
                {
                    oldState[i] = new RigidBody.State(bodies.GetBody(i));
                    bodies.GetBody(i).UpdateVel(dt);
                    bodies.GetBody(i).UpdatePos(dt);
                }

                // find all collision pairs
                collisionPairs.Clear();
                for (int i = 0; i < bodies.GetNumBodies(); i++)
                {
                    for (int j = i+1; j < bodies.GetNumBodies(); j++)
                    {
                        // TODO: better macro collision detection
                        CollisionPair cp = CheckCollision(new RigidBody[] {bodies.GetBody(i), bodies.GetBody(j)});
                        if (cp != null) collisionPairs.Add(cp);
                    }
                }

                // restore old states
                for (int i = 0; i < bodies.GetNumBodies(); i++)
                {
                    bodies.GetBody(i).SetState(oldState[i]);
                }

                // process all collision pairs
                foreach (CollisionPair cp in collisionPairs)
                {
                    cp.ResolveCollision(dt);
                }
            }
        }

        private void ProcessAllContacts(float dt)
        {
            if (num_contact_steps == 0 && !shock_propagation) return;

            // first, clear level and the collisions for each body
            for (int i = 0; i < bodies.GetNumBodies(); i++)
            {
                RigidBody a = bodies.GetBody(i);
                a.levelSet = false;
                a.collisions.Clear();
            }
                
            Queue queue = new Queue();
            for (int i = 0; i < bodies.GetNumBodies(); i++)
            {
                RigidBody a = bodies.GetBody(i);
                
                if (a.Mass == float.PositiveInfinity)
                {
                    // keep track of bodies w/ infinite mass for later
                    a.level = -1;
                    a.levelSet = true;
                    queue.Enqueue(a);

                    // these guys don't fall
                    continue;
                }
                
                // let each body fall due to gravity, check for collisions with all other bodies
                RigidBody.State savedState = new RigidBody.State(a);
                a.UpdatePos(dt);

                for (int j = 0; j < bodies.GetNumBodies(); j++)
                {
                    if (i == j) continue;
                    RigidBody b = bodies.GetBody(j);
                    CollisionPair cp = CheckCollision(new RigidBody[] {a, b});
                    if (cp != null) 
                    {
                        //Console.Out.WriteLine("contact between '{0}' and '{1}'", a.Name, b.Name);
                        a.collisions.Add(cp);
                        b.collisions.Add(cp);
                    }
                }

                a.SetState(savedState);
            }

            // topological sort (bfs)
            // TODO check this
            int max_level = -1;
            while (queue.Count > 0)
            {
                RigidBody a = queue.Dequeue() as RigidBody;
                //Console.Out.WriteLine("considering collisions with '{0}'", a.Name);
                if (a.level > max_level) max_level = a.level;
                foreach (CollisionPair cp in a.collisions)
                {
                    RigidBody b = (cp.body[0] == a ? cp.body[1] : cp.body[0]);
                    //Console.Out.WriteLine("considering collision between '{0}' and '{1}'", a.Name, b.Name);
                    if (!b.levelSet)
                    {
                        b.level = a.level + 1;
                        b.levelSet = true;
                        queue.Enqueue(b);
                        //Console.Out.WriteLine("found body '{0}' in level {1}", b.Name, b.level);
                    }
                }
            }

            int num_levels = max_level + 1;

            //Console.WriteLine("num_levels = {0}", num_levels);
            
            ArrayList[] bodiesAtLevel = new ArrayList[num_levels];
            ArrayList[] collisionsAtLevel = new ArrayList[num_levels];
            for (int i = 0; i < num_levels; i++)
            {
                bodiesAtLevel[i] = new ArrayList();
                collisionsAtLevel[i] = new ArrayList();
            }

            for (int i = 0; i < bodies.GetNumBodies(); i++)
            {
                RigidBody a = bodies.GetBody(i);
                if (!a.levelSet || a.level < 0) continue; // either a static body or no contacts

                // add a to a's level
                bodiesAtLevel[a.level].Add(a);
                
                // add collisions involving a to a's level
                foreach (CollisionPair cp in a.collisions)
                {
                    RigidBody b = (cp.body[0] == a ? cp.body[1] : cp.body[0]);
                    if (b.level <= a.level) // contact with object at or below the same level as a
                    {
                        // make sure not to add duplicate collisions
                        bool found = false;
                        foreach (CollisionPair cp2 in collisionsAtLevel[a.level])
                            if (cp == cp2) found = true;

                        if (!found) collisionsAtLevel[a.level].Add(cp);
                    }
                }
            }

            for (int step = 0; step < num_contact_steps; step++)
            {
                for (int level = 0; level < num_levels; level++)
                {
                    // process all contacts
                    foreach (CollisionPair cp in collisionsAtLevel[level])
                    {
                        cp.ResolveContact(dt, (num_contact_steps - step - 1) * -1.0f/num_contact_steps);
                    }
                }
            }

            if (shock_propagation)
            {
                for (int level = 0; level < num_levels; level++)
                {
                    // process all contacts
                    foreach (CollisionPair cp in collisionsAtLevel[level])
                    {
                        // inelastic collision
                        cp.ResolveContact(dt, 0);
                    }

                    // freeze each body in this level
                    foreach (RigidBody body in bodiesAtLevel[level])
                    {
                        body.frozen = true;
                    }
                }

                // unfreeze all bodies
                for (int level = 0; level < num_levels; level++)
                {
                    foreach (RigidBody body in bodiesAtLevel[level])
                    {
                        body.frozen = false;
                    }
                }
            }
        }


        public void Update(float dt)
        {
            if (dt == 0 || bodies == null) return;
            
            if (collisions_before_velocity_update) ProcessAllCollisions(dt);
  
            // update vel for each body
            for (int i = 0; i < bodies.GetNumBodies(); i++)
            {
                bodies.GetBody(i).UpdateVel(dt);
            }

            if (!collisions_before_velocity_update) ProcessAllCollisions(dt);
            ProcessAllContacts(dt);
            
            // update pos for each body
            for (int i = 0; i < bodies.GetNumBodies(); i++)
            {
                bodies.GetBody(i).UpdatePos(dt);
            }
        }
    }

    public class CollisionPair
    {
        public RigidBody[] body;
        
        // collisionPoints[i] are in the reference frame of body[i]
        public ArrayList[] collisionPoints;

        public CollisionPair(RigidBody[] b, ArrayList[] cp)
        {
            body = b;
            collisionPoints = cp;
        }

        const int max_iterations = 10; //HACK

        public void ResolveCollision(float dt)
        {
            //while (true)
            int iterations = 0;
            while (iterations++ < max_iterations) //HACK
            {
                // save body state, then update vel and pos for interference detection
                RigidBody.State[] curState = new RigidBody.State[2];
                for (int i = 0 ; i < 2; i++)
                {
                    curState[i] = new RigidBody.State(body[i]);
                    body[i].UpdateVel(dt);
                    body[i].UpdatePos(dt);
                }

                // find the deepest penetrating point
                SDM.Value bestCP = new SDM.Value(new Vector3(), new Vector3(), float.PositiveInfinity);
                for (int i = 0; i < 2; i++)
                {
                    for (int j = 0; j < collisionPoints[i].Count; )
                    {
                        SDM.Value cp = body[1-i].EvaluateSDM(body[1-i].PointWorldToBody(body[i].PointBodyToWorld((Vector3)collisionPoints[i][j])), true);
                        if (cp.d > 0)
                        {
                            collisionPoints[i].RemoveAt(j);
                            continue;
                        }

                        cp.p = body[1-i].PointBodyToWorld(cp.p);
                        cp.n = body[1-i].VectorBodyToWorld(cp.n);
                        if (i == 1) cp.n *= -1; // so don't have to worry about direction of normal
                        // remove any non-penetrating or separating points
                        //Vector3 vWorldRelative = curState[0].WorldVelocity(cp.p) - curState[1].WorldVelocity(cp.p);
                        Vector3 vWorldRelative = curState[0].vel + Vector3.Cross(curState[0].rotation, cp.p - curState[0].pos)
                            - (curState[1].vel + Vector3.Cross(curState[1].rotation, cp.p - curState[1].pos));
                        float vWorldRelativeN = Vector3.Dot(vWorldRelative, cp.n);
                        if (vWorldRelativeN >= -0.001) // HACK
                        {
                            collisionPoints[i].RemoveAt(j);
                            continue;
                        }

                        if (cp.d < bestCP.d)
                        {
                            bestCP = cp;
                        }

                        j++;
                    }
                }

                // restore original body state to do collision processing
                for (int i = 0; i < 2; i++)
                {
                    body[i].SetState(curState[i]);
                }
                
                if (collisionPoints[0].Count == 0 && collisionPoints[1].Count == 0)
                {
                    // no more points left
                    break;
                }

                // now have collision point and normal
                Vector3 p = bestCP.p;
                Vector3 N = bestCP.n;
                
                // constant of restitution ("bounciness")
                float r = Math.Min(body[0].Bounciness, body[1].Bounciness);
                
                GenerateImpulse(p, N, r, body[0], body[1]);
            }
        }

        public void ResolveContact(float dt, float r)
        {
            int iterations = 0;
            while (iterations++ < max_iterations) //HACK
            {
                // find the deepest penetrating point
                SDM.Value bestCP = new SDM.Value(new Vector3(), new Vector3(), float.PositiveInfinity);
                for (int i = 0; i < 2; i++)
                {
                    for (int j = 0; j < collisionPoints[i].Count; )
                    {
                        SDM.Value cp = body[1-i].EvaluateSDM(body[1-i].PointWorldToBody(body[i].PointBodyToWorld((Vector3)collisionPoints[i][j])), true);
                        if (cp.d > 0)
                        {
                            collisionPoints[i].RemoveAt(j);
                            continue;
                        }

                        cp.p = body[1-i].PointBodyToWorld(cp.p);
                        cp.n = body[1-i].VectorBodyToWorld(cp.n);
                        if (i == 1) cp.n *= -1; // so don't have to worry about direction of normal
                        // remove any non-penetrating points

                        Vector3 vWorldRelative = body[0].WorldVelocity(cp.p) - body[1].WorldVelocity(cp.p);
                        float vWorldRelativeN = Vector3.Dot(vWorldRelative, cp.n);
                        if (cp.d < bestCP.d && vWorldRelativeN < -0.001) // HACK
                        { // only choose points with non-separating velocity
                            bestCP = cp;
                        }

                        j++;
                    }
                }

                if (bestCP.d == float.PositiveInfinity)
                {
                    // all points are separating
                    break;
                }

                // now have collision point and normal
                Vector3 p = bestCP.p;
                Vector3 N = bestCP.n;
                
                GenerateImpulse(p, N, r, body[0], body[1]);
            }
        }


        private void GenerateImpulse(Vector3 p, Vector3 N, float r, RigidBody b0, RigidBody b1)
        {
            // frictional coefficient
            float u = Math.Max(body[0].Friction, body[1].Friction);

            Vector3 vOldWorldRelative = body[0].WorldVelocity(p) - body[1].WorldVelocity(p);
            float vOldWorldRelativeN = Vector3.Dot(vOldWorldRelative, N);
            Vector3 vNewWorldRelative = -r * vOldWorldRelativeN * N;

            Matrix3 K = body[0].ComputeK(p) + body[1].ComputeK(p);
            Vector3 J = -1 * K.Inverse() * (vNewWorldRelative - vOldWorldRelative);
            float Jn = Vector3.Dot(J,N);
            Vector3 Jt = J - (Jn * N);

            if (Jt.Length() >= u * Math.Abs(Jn))
            {
                // impulse not in friction cone; need to do sliding friction
                    
                Vector3 vOldWorldRelativeT = vOldWorldRelative - (vOldWorldRelativeN * N);
                float TLength = vOldWorldRelativeT.Length();
                Vector3 T = (TLength == 0 ? Vector3.Empty : vOldWorldRelativeT * (1 / TLength));

                Jn = ((r + 1) * vOldWorldRelativeN) / (Vector3.Dot(K * (N - (u * T)), N));
                J = Jn * (N - (u * T));
            }

            // XXX HACK
            //if (Math.Abs(Jn) < 0.001) break;

            // apply impulse to both bodies
            body[0].ApplyImpulse(p, -1 * J);
            body[1].ApplyImpulse(p, J);
        }

    }


}